use io;

/* Read instance data. The input files follow the format of the DIMACS challenge */
function input() {
    usage = "Usage: localsolver capacitated_arc_routing.lsp "
            + "inFileName=inputFile [solFileName=outputFile] [lsTimeLimit=timeLimit]";

    if (inFileName == nil) throw usage;
    readData();
}

/* Declare the optimization model */
function model() {
    // Sequence of edges visited and "serviced" by each truck
    // A sequence can contain an edge in one direction or its reverse
    edgesSequences[k in 0...nbTrucks] <- list(2 * nbRequiredEdges);

    // An edge must be serviced by at most one truck
    constraint disjoint(edgesSequences);

    // An edge can be travelled in both directions but its demand must be
    // satisfied only once
    for [i in 0...nbRequiredEdges] {
        constraint contains(edgesSequences, 2 * i) + contains(edgesSequences, 2 * i + 1) == 1;
    }

    for [k in 0...nbTrucks] {
        local sequence <- edgesSequences[k];
        local c <- count(sequence);

        // Quantity in each truck
        routeQuantity <- sum(sequence, j => requiredEdgesDemands[j]);
        // Capacity constraint : a truck must not exceed its capacity
        constraint routeQuantity <= truckCapacity;

        // Distance travelled by each truck
        routeDistance[k] <- sum(1...c, i => requiredEdgesCosts[sequence[i]] +
                distanceBetweenEdges[sequence[i - 1]][sequence[i]]) +
                (c > 0 ? requiredEdgesCosts[sequence[0]] + distanceFromDepot[sequence[0]] +
                distanceToDepot[sequence[c - 1]] : 0);
    }

    // Total distance travelled
    totalDistance <- sum[k in 0...nbTrucks](routeDistance[k]);

    // Objective: minimize the distance travelled
    minimize totalDistance;
}

/* Parametrize the solver */
function param() {
    if (lsTimeLimit == nil) lsTimeLimit = 20;
}

/* Write the solution in a file with the following format:
 * - total distance
 * - number of routes
 * - for each truck, the edges visited */
function output() {
    if (solFileName == nil) return;
    local outfile = io.openWrite(solFileName);
    outfile.println("Objective function value : ", totalDistance.value);
    outfile.println("Number of routes : ", nbTrucks);
    for [k in 0...nbTrucks] {
        outfile.print("Sequence of truck ", k + 1, ": ");
        local sequence = edgesSequences[k].value;
        for [i in 0...sequence.count()] {
            outfile.print("(", requiredEdgesOrigin[sequence[i]], 
                    ", ", requiredEdgesDestination[sequence[i]], ")  ");
        }
        outfile.println();
    }
}

function readData() {
    local inFile = io.openRead(inFileName);
    for [_ in 0...2] inFile.readln();
    nbNodes = inFile.readln().trim().split(":")[1].toInt();
    nbRequiredEdges = inFile.readln().trim().split(":")[1].toInt();
    nbNotRequiredEdges = inFile.readln().trim().split(":")[1].toInt();
    nbTrucks = inFile.readln().trim().split(":")[1].toInt();
    truckCapacity = inFile.readln().trim().split(":")[1].toInt();
    for [_ in 0...2] inFile.readln();

    inFile.readln();
    requiredNodes[0] = -1;
    for [i in 0...nbNodes][j in 0...nbNodes] nodeNeighbors[i][j] = 0;
    for [i in 0...nbRequiredEdges] {
        local elements = inFile.readln().trim().split("   ");
        local firstNode = elements[0].trim().split(",")[0].trim().split()[1].toInt();
        local secondNode = elements[0].trim().split(",")[1].trim().replace(")", "").toInt();
        local cost = elements[1].trim().split()[1].toInt();
        local demand = elements[2].trim().split()[1].toInt();
        // Even indices store direct edges, and odd indices store reverse edges
        requiredEdgesCosts[2 * i] = cost;
        requiredEdgesDemands[2 * i] = demand;
        requiredEdgesCosts[2 * i + 1] = cost;
        requiredEdgesDemands[2 * i + 1] = demand;
        requiredEdgesOrigin[2 * i] = firstNode;
        requiredEdgesDestination[2 * i] = secondNode;
        requiredEdgesOrigin[2 * i + 1] = secondNode;
        requiredEdgesDestination[2 * i + 1] = firstNode;
        addRequiredNodes(firstNode);
        addRequiredNodes(secondNode);
        nodeNeighbors[firstNode - 1][secondNode - 1] = cost;
        nodeNeighbors[secondNode - 1][firstNode - 1] = cost;
    }

    if (nbNotRequiredEdges > 0) {
        inFile.readln();
        for [i in 0...nbNotRequiredEdges] {
            local elements = inFile.readln().trim().split("   ");
            local firstNode = elements[0].trim().split(",")[0].trim().split()[1].toInt();
            local secondNode = elements[0].trim().split(",")[1].trim().replace(")", "").toInt();
            local cost = elements[1].trim().split()[1].toInt();
            nodeNeighbors[firstNode - 1][secondNode - 1] = cost;
            nodeNeighbors[secondNode - 1][firstNode - 1] = cost;
        }
    }
    depotNode = inFile.readln().trim().split(":")[1].toInt();
    nbRequiredNodes = requiredNodes.count();
    for [i in 0...nbNodes] {
        shortestPath[i] = 0;
    }
    findRequiredPaths(shortestPath);
    findDistanceBetweenEdges();
    findDistanceToDepot();
    findDistanceFromDepot(shortestPath);
}

/* Finds the shortest path from one node "origin" to all the other nodes of the graph
   thanks to the Dijkstra's algorithm */
function minDistance(shortestPath, sptSet) {
    minValue = +inf;
    minIndex = -1;
    for [i in 0...nbNodes] {
        if (sptSet[i] == false && shortestPath[i] <= minValue) {
            minValue = shortestPath[i];
            minIndex = i;
        }
    }
    return minIndex;
}

function shortestPathFinder(origin, shortestPath) {
    local positiveInfinity =  +inf;
    for [i in 0...nbNodes] {
        shortestPath[i] = positiveInfinity;
        sptSet[i] = false;
    }
    shortestPath[origin - 1] = 0;
    for [_ in 0...nbNodes] {
        currentNode = minDistance(shortestPath, sptSet);
        sptSet[currentNode] = true;
        currentNeighbors = nodeNeighbors[currentNode];
        for [neighbor in 0...nbNodes] {
            if (currentNeighbors[neighbor] != 0) {
                distance = currentNeighbors[neighbor];
                if (!sptSet[neighbor] &&
                        shortestPath[currentNode] + distance < shortestPath[neighbor]) {
                    shortestPath[neighbor] = shortestPath[currentNode] + distance;
                }
            }
        }
    }
}

function findRequiredPaths(shortestPath) {
    for [i in 0...nbRequiredNodes] {
        node = requiredNodes[i];
        shortestPathFinder(node, shortestPath);
        for [j in 0...nbNodes] {
            requiredDistance[i][j] = shortestPath[j];
        }
    }
}

function findDistanceBetweenEdges() {
    for [i in 0...2*nbRequiredEdges] {
        for [j in 0...2*nbRequiredEdges] {
            if (requiredEdgesDestination[i] == requiredEdgesOrigin[j]) {
                distanceBetweenEdges[i][j] = 0;
            } else {
                for [k in 0...nbRequiredNodes] {
                    if (requiredNodes[k] == requiredEdgesDestination[i]) {
                        distanceBetweenEdges[i][j] = requiredDistance[k][requiredEdgesOrigin[j] - 1];
                    }
                }
            }
        }
    }
}

function findDistanceToDepot() {
    for [i in 0...2*nbRequiredEdges] {
        if (requiredEdgesDestination[i] == depotNode) {
            distanceToDepot[i] = 0;
        } else {
            for [k in 0...nbRequiredNodes] {
                if (requiredNodes[k] == requiredEdgesDestination[i]) {
                    distanceToDepot[i] = requiredDistance[k][depotNode - 1];
                }
            }
        }
    }
}

function findDistanceFromDepot(shortestPath) {
    for [i in 0...2*nbRequiredEdges] {
        if (depotNode == requiredEdgesOrigin[i]) {
            distanceFromDepot[i] = 0;
        } else {
            depotIsRequiredNode = false;
            for [k in 0...nbRequiredNodes] {
                if (requiredNodes[k] == depotNode) {
                    depotIsRequiredNode = true;
                    distanceFromDepot[i] = requiredDistance[k][requiredEdgesOrigin[i] - 1];
                }
            }
            if (!depotIsRequiredNode) {
                shortestPathFinder(depotNode, shortestPath);
                distanceFromDepot[i] = shortestPath[requiredEdgesOrigin[i] - 1];
            }
        }
    }
}

function addRequiredNodes(node) {
    local n = requiredNodes.count();
    local found = false;
    for [i in 0...n] {
        if (requiredNodes[i] == node) {
            found = true;
        }
    }
    if (!found) {
        if (requiredNodes[0] == -1) {
            requiredNodes[0] = node;
        } else {
            requiredNodes[n] = node;
        }
    }
}
